// SPDX-License-Identifier: GPL-2.0-only
/*
 * Copyright (C) 2022 Unfolded Circle
 * Author: Marton Borzak <marton@unfoldedcircle.com>
 */

#include <linux/backlight.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/of_device.h>

#include <linux/gpio/consumer.h>
#include <linux/regulator/consumer.h>

#include <drm/drm_mipi_dsi.h>
#include <drm/drm_modes.h>
#include <drm/drm_panel.h>
#include <drm/drm_crtc.h>
#include <linux/media-bus-format.h>

#include "ams326pm01_gamma.h"

#define MIN_BRIGHTNESS		0
#define MAX_BRIGHTNESS		10

struct ams326pm01 {
	struct drm_panel		panel;
	struct mipi_dsi_device	*dsi;
	struct device 			*dev;
	struct backlight_device *backlight;

	struct gpio_desc		*enable_gpio;

	bool 					enabled;
	struct gpio_desc		*cs_gpio;
	struct gpio_desc		*sclk_gpio;
	struct gpio_desc		*sdi_gpio;
	struct gpio_desc		*reset_gpio;
};

/* * * * * * * * * * * * * * * * * * * ICN6211 FUNCTIONS * * * * * * * * * * * * * * * * * * */
struct ICN6211_INIT_CMD {
	size_t len;
	const char *data;
};

#define ICN6211_INIT_CMD(...) { \
	.len = sizeof((char[]){__VA_ARGS__}), \
	.data = (char[]){__VA_ARGS__} }

static const struct ICN6211_INIT_CMD ICN6211_INIT_CMDs[] = {
	ICN6211_INIT_CMD(0x7A, 0xC1),
	ICN6211_INIT_CMD(0x18, 0x00),
	ICN6211_INIT_CMD(0x19, 0x00),
	ICN6211_INIT_CMD(0x1A, 0x00),
	ICN6211_INIT_CMD(0x1B, 0x00),
	ICN6211_INIT_CMD(0x20, 0x56),
	ICN6211_INIT_CMD(0x21, 0xE0),
	ICN6211_INIT_CMD(0x22, 0x13),
	ICN6211_INIT_CMD(0x23, 0x19),
	ICN6211_INIT_CMD(0x24, 0x02),
	ICN6211_INIT_CMD(0x25, 0x7A),
	ICN6211_INIT_CMD(0x26, 0x00),
	ICN6211_INIT_CMD(0x27, 0x16),
	ICN6211_INIT_CMD(0x28, 0x02),
	ICN6211_INIT_CMD(0x29, 0x26),
	ICN6211_INIT_CMD(0x34, 0x80),
	ICN6211_INIT_CMD(0x36, 0x19),
	ICN6211_INIT_CMD(0x86, 0x29),
	ICN6211_INIT_CMD(0xB5, 0xA0),
	ICN6211_INIT_CMD(0x5C, 0xFF),
	ICN6211_INIT_CMD(0x2A, 0x07),
	ICN6211_INIT_CMD(0x56, 0x92),
	ICN6211_INIT_CMD(0x59, 0xE1),
	ICN6211_INIT_CMD(0x6B, 0x72),
	ICN6211_INIT_CMD(0x69, 0x27),
	ICN6211_INIT_CMD(0x10, 0x50),
	ICN6211_INIT_CMD(0x11, 0x98),
	ICN6211_INIT_CMD(0xB6, 0x20),
	ICN6211_INIT_CMD(0x51, 0x20),
	ICN6211_INIT_CMD(0x09, 0x10),

	/* Test mode */
	// ICN6211_INIT_CMD(0x14, 0x43),
	// ICN6211_INIT_CMD(0x2a, 0x4e),
};

static int icn6211_reset(struct ams326pm01 *ctx)
{
	gpiod_set_value_cansleep(ctx->enable_gpio, 0);
	msleep(50);
	gpiod_set_value_cansleep(ctx->enable_gpio, 1);

	dev_info(ctx->dev, "ICN6211 reset\n");

	return 0;
}

static int icn6211_poweroff(struct ams326pm01 *ctx)
{
	gpiod_set_value_cansleep(ctx->enable_gpio, 0);

	dev_info(ctx->dev, "ICN6211 poweroff\n");
	
	return 0;
}

static int icn6211_init(struct ams326pm01 *ctx)
{
	struct mipi_dsi_device *dsi = ctx->dsi;
	unsigned int i; 
	int ret;

	for (i = 0; i < ARRAY_SIZE(ICN6211_INIT_CMDs); i++) {
		const struct ICN6211_INIT_CMD *cmd = &ICN6211_INIT_CMDs[i];

		ret = mipi_dsi_generic_write(dsi, cmd->data, cmd->len);
		if (ret < 0) {
			dev_err(ctx->dev, "DSI write error: %d\n", ret);
			return ret;
		}
	}

	dev_info(ctx->dev, "ICN6211 init\n");

	return 0;
}

/* * * * * * * * * * * * * * * * * * * DISPLAY FUNCTIONS * * * * * * * * * * * * * * * * * * */
static void ams326pm01_spi_write_byte(struct ams326pm01 *lcd, u8 cmd)
{
	u32 kk;

	gpiod_set_value_cansleep(lcd->sclk_gpio, 0);

	for (kk = 0; kk<8; kk++)
	{
		if (cmd&0x80) {
			gpiod_set_value_cansleep(lcd->sdi_gpio, 1);
		} else {
			gpiod_set_value_cansleep(lcd->sdi_gpio, 0);
		}
		cmd<<=1;
		gpiod_set_value_cansleep(lcd->sclk_gpio, 0);
		ndelay(50);
		gpiod_set_value_cansleep(lcd->sclk_gpio, 1);
		ndelay(50);
	}
};

static void ams326pm01_spi_write_cmd(struct ams326pm01 *lcd, u8 addr, u8 value)
{
	gpiod_set_value_cansleep(lcd->cs_gpio, 0);
	ams326pm01_spi_write_byte(lcd, addr);
	ams326pm01_spi_write_byte(lcd, value);
	gpiod_set_value_cansleep(lcd->cs_gpio, 1);
};

static void ams326pm01_display_init(struct ams326pm01 *lcd)
{
	/* panel condition set */
	ams326pm01_spi_write_cmd(lcd,0x27,0x02);
	ams326pm01_spi_write_cmd(lcd,0x28,0x88);
	ams326pm01_spi_write_cmd(lcd,0x29,0x65);
	ams326pm01_spi_write_cmd(lcd,0x2a,0x90);
	ams326pm01_spi_write_cmd(lcd,0x2b,0x55);
	ams326pm01_spi_write_cmd(lcd,0x2c,0x21);
	ams326pm01_spi_write_cmd(lcd,0x2d,0x31);
	ams326pm01_spi_write_cmd(lcd,0x2e,0x10);
	ams326pm01_spi_write_cmd(lcd,0x2f,0x41);
	ams326pm01_spi_write_cmd(lcd,0x30,0x21);
	ams326pm01_spi_write_cmd(lcd,0x36,0x05);
	ams326pm01_spi_write_cmd(lcd,0x37,0x07);
	ams326pm01_spi_write_cmd(lcd,0x38,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x39,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x3a,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x3b,0x1c);

	/* display condition set */
	ams326pm01_spi_write_cmd(lcd,0x05,0x0f);
	ams326pm01_spi_write_cmd(lcd,0x15,0x00);
	ams326pm01_spi_write_cmd(lcd,0x19,0x03);
	ams326pm01_spi_write_cmd(lcd,0x1a,0x56);
	ams326pm01_spi_write_cmd(lcd,0x1d,0x26);  
	ams326pm01_spi_write_cmd(lcd,0x1f,0x7a);  
	ams326pm01_spi_write_cmd(lcd,0x3c,0x57);
	ams326pm01_spi_write_cmd(lcd,0x3e,0x16);
	ams326pm01_spi_write_cmd(lcd,0x3f,0x0c);
	ams326pm01_spi_write_cmd(lcd,0x40,0x06);
	ams326pm01_spi_write_cmd(lcd,0x41,0x06);
	ams326pm01_spi_write_cmd(lcd,0x42,0x06);

	/* pentile gamma condition set */
	ams326pm01_spi_write_cmd(lcd,0xff,0x01);
	ams326pm01_spi_write_cmd(lcd,0x0a,0x12);
	ams326pm01_spi_write_cmd(lcd,0x0b,0x12);
	ams326pm01_spi_write_cmd(lcd,0x0c,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x0d,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x0e,0x1b);
	ams326pm01_spi_write_cmd(lcd,0x0f,0x13);
	ams326pm01_spi_write_cmd(lcd,0x10,0x24);
	ams326pm01_spi_write_cmd(lcd,0x11,0x21);
	ams326pm01_spi_write_cmd(lcd,0x2b,0x12);
	ams326pm01_spi_write_cmd(lcd,0x2c,0x12);
	ams326pm01_spi_write_cmd(lcd,0x2d,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x2e,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x2f,0x1b);
	ams326pm01_spi_write_cmd(lcd,0x30,0x13);
	ams326pm01_spi_write_cmd(lcd,0x31,0x24);
	ams326pm01_spi_write_cmd(lcd,0x32,0x21);
	ams326pm01_spi_write_cmd(lcd,0x4c,0x12);
	ams326pm01_spi_write_cmd(lcd,0x4d,0x12);
	ams326pm01_spi_write_cmd(lcd,0x4e,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x4f,0x1c);
	ams326pm01_spi_write_cmd(lcd,0x50,0x1b);
	ams326pm01_spi_write_cmd(lcd,0x51,0x13);
	ams326pm01_spi_write_cmd(lcd,0x52,0x24);
	ams326pm01_spi_write_cmd(lcd,0x53,0x21);
	ams326pm01_spi_write_cmd(lcd,0xff,0x00);

	ndelay(100);	

	dev_info(lcd->dev, "Display init commands sent\n");
};

static void ams326pm01_sleep_in(struct ams326pm01 *lcd)
{
	dev_info(lcd->dev, "Display sleep in\n");
	ams326pm01_spi_write_cmd(lcd, 0x00, 0xA1);
};

static void ams326pm01_sleep_out(struct ams326pm01 *lcd)
{
	ams326pm01_spi_write_cmd(lcd, 0x00, 0xA0);
	dev_info(lcd->dev, "Display sleep out\n");
};

static void _ams326pm0_gamma_ctl(struct ams326pm01 *lcd, const unsigned int *gamma)
{
	unsigned int i = 0;
	
	/* gamma set update disable command */
	ams326pm01_spi_write_cmd(lcd, 0x93, 0x01);

	for (i = 0; i < GAMMA_TABLE_COUNT; i++) {
		ams326pm01_spi_write_cmd(lcd, ams326pm01_gamma_commands[i], gamma[i]);
	}

	/* gamma set update enable command */
	ams326pm01_spi_write_cmd(lcd, 0x93, 0x00);
};

static void ams326pm0_gamma_ctl(struct ams326pm01 *lcd, int brightness)
{
	_ams326pm0_gamma_ctl(lcd, gamma_table.gamma_table[brightness]);
};

static int ams326pm01_prepare_gpios(struct ams326pm01 *ctx)
{
	int ret;

	ret = gpiod_direction_output(ctx->cs_gpio, 1);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for cs.\n");
		return ret;
	}
	ret = gpiod_direction_output(ctx->sclk_gpio, 0);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for clk.\n");
		return ret;
	}
	ret = gpiod_direction_output(ctx->sdi_gpio, 1);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for data.\n");
		return ret;
	}

	return 0;
};

static int ams326pm01_unprepare_gpios(struct ams326pm01 *ctx)
{
	int ret;

	ret = gpiod_direction_input(ctx->cs_gpio);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for cs.\n");
		return ret;
	}
	ret = gpiod_direction_input(ctx->sclk_gpio);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for clk.\n");
		return ret;
	}
	ret = gpiod_direction_input(ctx->sdi_gpio);
	if (ret < 0) {
		dev_err(ctx->dev, "GPIO setup failed for data.\n");
		return ret;
	}

	return 0;
};

static void ams326pm01_reset(struct ams326pm01 *lcd)
{
	gpiod_set_value_cansleep(lcd->reset_gpio, 0);
	msleep(10);
	gpiod_set_value_cansleep(lcd->reset_gpio, 1);
	msleep(10);	
	dev_info(lcd->dev, "Display reset\n");
};

static int ams326pm01_power_on(struct ams326pm01 *lcd)
{
	ams326pm01_prepare_gpios(lcd);

	ams326pm01_reset(lcd);
	ams326pm01_display_init(lcd);
	ams326pm0_gamma_ctl(lcd, 9); // max value is 10
	msleep(50);
	ams326pm01_sleep_out(lcd);

	lcd->enabled = true;

	dev_info(lcd->dev, "Display power on\n");

	return 0;
}

static int ams326pm01_power_off(struct ams326pm01 *lcd)
{
	ams326pm01_sleep_in(lcd);
	
	msleep(250);
	
	gpiod_set_value_cansleep(lcd->reset_gpio, 0);

	lcd->enabled = false;

	ams326pm01_unprepare_gpios(lcd);

	dev_info(lcd->dev, "Display power off\n");

	return 0;
}

static int  ams326pm01_on_sequence(struct ams326pm01 *ctx)
{
	if (ctx->enabled) {
		dev_info(ctx->dev, "On sequence, nothing to do.\n");
		return 0;
	}

	struct mipi_dsi_device *dsi = ctx->dsi;
	unsigned int i;
	int ret;

	dsi->mode_flags |= MIPI_DSI_MODE_LPM;

	icn6211_reset(ctx);
	msleep(10);
	icn6211_init(ctx);

	ret = mipi_dsi_dcs_exit_sleep_mode(dsi);
	if (ret < 0) {
		dev_err(ctx->dev, "Failed to exit sleep mode: %d\n", ret);
		return ret;
	}

	msleep(20);

	ret = mipi_dsi_dcs_set_display_on(dsi);
	if (ret < 0) {
		dev_err(ctx->dev, "Failed to set display on: %d\n", ret);
		return ret;
	}

	ams326pm01_power_on(ctx);

	dev_info(ctx->dev, "On sequence done.\n");

	return 0;
}

static int  ams326pm01_off_sequence(struct ams326pm01 *ctx)
{
	if (!ctx->enabled) {
		dev_info(ctx->dev, "Off sequence, nothing to do.\n");
		return 0;
	}

	struct mipi_dsi_device *dsi = ctx->dsi;
	int ret;

	ams326pm01_power_off(ctx);

	dsi->mode_flags &= ~MIPI_DSI_MODE_LPM;

	ret = mipi_dsi_dcs_set_display_off(dsi);
	if (ret < 0) {
		dev_err(ctx->dev, "Failed to set display off: %d\n", ret);
		return ret;
	}

	ret = mipi_dsi_dcs_enter_sleep_mode(dsi);
	if (ret < 0) {
		dev_err(ctx->dev, "Failed to enter sleep mode: %d\n", ret);
		return ret;
	}
	msleep(120);

	icn6211_poweroff(ctx);

	msleep(50);

	dev_info(ctx->dev, "Off sequence done.\n");

	return 0;
}

static int ams326pm01_set_brightness(struct backlight_device *bd)
{
	struct mipi_dsi_device *dsi = bl_get_data(bd);
	struct ams326pm01 *lcd = mipi_dsi_get_drvdata(dsi);
	int brightness = bd->props.brightness;
	
	if (brightness < MIN_BRIGHTNESS ||
		brightness > bd->props.max_brightness) {
		dev_err(&bd->dev, "lcd brightness should be %d to %d.\n",
			MIN_BRIGHTNESS, MAX_BRIGHTNESS);
		return -EINVAL;
	}

	if (brightness == 0 && lcd->enabled) {
		ams326pm01_off_sequence(lcd);
	} else if (brightness > 0 && !lcd->enabled) {
		ams326pm01_on_sequence(lcd);
	} else {
		ams326pm0_gamma_ctl(lcd, brightness);
	}

	dev_info(lcd->dev, "Display brightness changed: %i\n", brightness);

	return 0;
}

static const struct backlight_ops ams326pm01_backlight_ops  = {
	.update_status = ams326pm01_set_brightness,
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static inline struct ams326pm01 *panel_to_ams326pm01(struct drm_panel *panel)
{
	return container_of(panel, struct ams326pm01, panel);
}

static int ams326pm01_prepare(struct drm_panel *panel)
{
	struct ams326pm01 *ctx = panel_to_ams326pm01(panel);
	
	ams326pm01_on_sequence(ctx);

	return 0;
}

static int ams326pm01_unprepare(struct drm_panel *panel)
{
	struct ams326pm01 *ctx = panel_to_ams326pm01(panel);

	ams326pm01_off_sequence(ctx);

	return 0;
}

static const struct drm_display_mode ams326pm01_default_mode = {
	.clock = 32400,

	.hdisplay = 854,
	.hsync_start = 854 + 122,
	.hsync_end = 854 + 122 + 2,
	.htotal = 854 + 122 + 2 + 25,

	.vdisplay = 480,
	.vsync_start = 480 + 22,
	.vsync_end = 480 + 22 + 2,
	.vtotal = 480 + 22 + 2 + 38,

	.width_mm = 80,
	.height_mm = 47,
	
	.flags = DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC,
};

static int ams326pm01_get_modes(struct drm_panel *panel, struct drm_connector *connector)
{
	struct drm_display_mode *mode;

	mode = drm_mode_duplicate(connector->dev, &ams326pm01_default_mode);
	if (!mode) {
		dev_err(panel->dev, "failed to add mode %ux%ux@%u\n",
			ams326pm01_default_mode.hdisplay,
			ams326pm01_default_mode.vdisplay,
			drm_mode_vrefresh(&ams326pm01_default_mode));
		return -ENOMEM;
	}

	drm_mode_set_name(mode);

	connector->display_info.width_mm = mode->width_mm;
	connector->display_info.height_mm = mode->height_mm;
	connector->display_info.bus_flags = DRM_BUS_FLAG_PIXDATA_DRIVE_POSEDGE;
	
	static const u32 bus_format = MEDIA_BUS_FMT_RGB888_1X24;
	drm_display_info_set_bus_formats(&connector->display_info, &bus_format, 1);

	mode->type = DRM_MODE_TYPE_DRIVER | DRM_MODE_TYPE_PREFERRED;
	drm_mode_probed_add(connector, mode);

	return 1;
}

static const struct drm_panel_funcs ams326pm01_funcs = {
	.unprepare = ams326pm01_unprepare,
	.prepare = ams326pm01_prepare,
	.get_modes = ams326pm01_get_modes,
};

static int ams326pm01_dsi_probe(struct mipi_dsi_device *dsi)
{
	struct ams326pm01 *ctx;
	struct backlight_properties bl_props;
	int ret;

	ctx = devm_kzalloc(&dsi->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	ctx->enable_gpio = devm_gpiod_get(&dsi->dev, "enable", GPIOD_OUT_LOW);
	if (IS_ERR(ctx->enable_gpio)) {
		dev_err(&dsi->dev, "Couldn't get our enable GPIO\n");
		return PTR_ERR(ctx->enable_gpio);
	}

	// display gpios
	ctx->reset_gpio = devm_gpiod_get(&dsi->dev, "reset", GPIOD_OUT_HIGH);
	if (IS_ERR(ctx->reset_gpio)) {
		dev_err(&dsi->dev, "failed to request reset GPIO\n");
		return PTR_ERR(ctx->reset_gpio);
	}
	ctx->cs_gpio = devm_gpiod_get(&dsi->dev, "cs", GPIOD_OUT_HIGH);
	if (IS_ERR(ctx->cs_gpio)) {
		dev_err(&dsi->dev, "failed to request cs GPIO\n");
		return PTR_ERR(ctx->cs_gpio);
	}
	ctx->sclk_gpio = devm_gpiod_get(&dsi->dev, "clk", GPIOD_OUT_HIGH);
	if (IS_ERR(ctx->sclk_gpio)) {
		dev_err(&dsi->dev, "failed to request clk GPIO\n");
		return PTR_ERR(ctx->sclk_gpio);
	}
	ctx->sdi_gpio = devm_gpiod_get(&dsi->dev, "data", GPIOD_OUT_HIGH);
	if (IS_ERR(ctx->sdi_gpio)) {
		dev_err(&dsi->dev, "failed to request data GPIO\n");
		return PTR_ERR(ctx->sdi_gpio);
	}

	ctx->dsi = dsi;
	ctx->dev = &dsi->dev;

	mipi_dsi_set_drvdata(dsi, ctx);

	dsi->format = MIPI_DSI_FMT_RGB888;
	dsi->mode_flags = MIPI_DSI_MODE_VIDEO | MIPI_DSI_MODE_VIDEO_BURST | MIPI_DSI_MODE_NO_EOT_PACKET;
	dsi->lanes = 2;

	memset(&bl_props, 0, sizeof(bl_props));
	bl_props.type = BACKLIGHT_RAW;
	bl_props.brightness = MAX_BRIGHTNESS;
	bl_props.max_brightness = MAX_BRIGHTNESS;

	ctx->backlight = devm_backlight_device_register(ctx->dev, dev_name(ctx->dev),
							  ctx->dev, dsi, &ams326pm01_backlight_ops,
							  &bl_props);
	if (IS_ERR(ctx->backlight)) {
		ret = PTR_ERR(ctx->backlight);
		dev_err(ctx->dev, "Failed to register backlight (%d)\n", ret);
		return ret;
	}

	drm_panel_init(&ctx->panel, &dsi->dev, &ams326pm01_funcs, DRM_MODE_CONNECTOR_DSI);
	ctx->panel.dev = &dsi->dev;
	ctx->panel.funcs = &ams326pm01_funcs;

	drm_panel_add(&ctx->panel);
	
	ret = mipi_dsi_attach(dsi);
	if (ret < 0) {
		dev_err(ctx->dev, "Failed to attach to DSI host: %d\n", ret);
		drm_panel_remove(&ctx->panel);
		return ret;
	}

	return 0;
}

static void ams326pm01_dsi_remove(struct mipi_dsi_device *dsi)
{
	struct ams326pm01 *ctx = mipi_dsi_get_drvdata(dsi);
	int ret;

	ret = mipi_dsi_detach(dsi);
	if (ret < 0)
		dev_err(&dsi->dev, "Failed to detach from DSI host: %d\n", ret);

	drm_panel_remove(&ctx->panel);
}

static const struct of_device_id ams326pm01_of_match[] = {
	{ .compatible = "uc,ams326pm01-icn6211", },
	{ }
};

MODULE_DEVICE_TABLE(of, ams326pm01_of_match);

static struct mipi_dsi_driver ams326pm01_driver = {
	.probe = ams326pm01_dsi_probe,
	.remove = ams326pm01_dsi_remove,
	.driver = {
		.name = "uc-ams326pm01-icn6211",
		.of_match_table = ams326pm01_of_match,
	},
};
module_mipi_dsi_driver(ams326pm01_driver);

MODULE_AUTHOR("Marton Borzak <marton@unfoldedcircle.com>");
MODULE_DESCRIPTION("UC OLED DISPLAY ICN6211 MIPI-DSI to RGB");
MODULE_LICENSE("GPL v2");
