// SPDX-License-Identifier: GPL-2.0-only
/*
 * MC3419 accelerometer driver
 *
 * Copyright (C) 2022 Marton Borzak <marton@unfoldedcircle.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 */

#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/regmap.h>
#include <linux/slab.h>
#include <linux/input.h>

#define MC3419_DEVICE_ID			0x6c
#define MC3419_REG_MAX				0x4b

#define MC3419_REGMAP_NAME			"mc3419_regmap"
#define MC3419_REG_STATUS			0x05
#define MC3419_REG_MODE				0x07
#define MC3419_REG_INT_ENABLE		0x06
#define MC3419_REG_MOTION_CTRL		0x09
#define MC3419_REG_INT_CLEAR		0x14
#define MC3419_REG_RANGE			0x20
#define MC3419_REG_SAMPLE_RATE		0x08
#define MC3419_REG_INT_SWAP			0x31
#define MC3419_REG_GPIO				0x33
#define MC3419_REG_ANYM_TRESHOLD_1	0x43
#define MC3419_REG_ANYM_TRESHOLD_2	0x44

struct mc3419_data {
	struct device *dev;
	struct regmap *regmap;
	struct i2c_client *client;
	struct input_dev  *input;
	int anymotion_threshold;
	int interrupt_enabled;
};

static const struct i2c_device_id mc3419_id[] = {
	{ "mc3419", MC3419_DEVICE_ID },
	{}
};

MODULE_DEVICE_TABLE(i2c, mc3419_id);

static const struct of_device_id mc3419_of_table[] = {
	{ .compatible = "mcube,mc3419" },
	{}
};

MODULE_DEVICE_TABLE(of, mc3419_of_table);

static int mc3419_read_reg(struct mc3419_data *mc, u8 reg, unsigned int *val)
{
	return regmap_read(mc->regmap, reg, val);
}

static int mc3419_write_reg(struct mc3419_data *mc, u8 reg, u8 val)
{
	return regmap_write(mc->regmap, reg, val);
}

static irqreturn_t mc3419_irq_handler(int irq, void *data)
{
	struct mc3419_data *mc = data;

	int ret = 0;
	unsigned int value = -1;

	ret = mc3419_read_reg(mc, MC3419_REG_INT_CLEAR, &value);
	if (ret < 0)
		dev_err(mc->dev, "Cannot read interrupt register.\n");

	dev_dbg(mc->dev, "Interrupt register: %i\n", value);

	// clear interrupt here
	ret = mc3419_write_reg(mc, MC3419_REG_INT_CLEAR, 0x00);
	if (ret < 0)
		dev_err(mc->dev, "Cannot clear interrupt.\n");

	dev_dbg(mc->dev, "Accelerometer interrupt.\n");

	input_report_key(mc->input, KEY_WAKEUP, 0);
	input_report_key(mc->input, KEY_WAKEUP, 1);
	input_sync(mc->input);

	return IRQ_HANDLED;
}

static bool mc3419_is_readable_reg(struct device *dev, unsigned int reg)
{
	switch (reg) {
	case MC3419_REG_STATUS:
	case MC3419_REG_INT_CLEAR:
	case MC3419_REG_RANGE:
	case MC3419_REG_SAMPLE_RATE:
		return true;
	default:
		return false;
	}
}

static bool mc3419_is_writeable_reg(struct device *dev, unsigned int reg)
{
	switch (reg) {
	case MC3419_REG_MODE:
	case MC3419_REG_INT_ENABLE:
	case MC3419_REG_MOTION_CTRL:
	case MC3419_REG_INT_CLEAR:
	case MC3419_REG_RANGE:
	case MC3419_REG_SAMPLE_RATE:
	case MC3419_REG_INT_SWAP:
	case MC3419_REG_GPIO:
	case MC3419_REG_ANYM_TRESHOLD_1:
	case MC3419_REG_ANYM_TRESHOLD_2:
		return true;
	default:
		return false;
	}
}

static const struct regmap_config mc3419_regmap_config = {
	.name = MC3419_REGMAP_NAME,

	.reg_bits = 8,
	.val_bits = 8,

	.max_register =  MC3419_REG_MAX,

	.readable_reg = mc3419_is_readable_reg,
	.writeable_reg = mc3419_is_writeable_reg,
};

static int mc3419_standby(struct mc3419_data *mc) {
	int ret = 0;

	ret = mc3419_write_reg(mc, MC3419_REG_MODE, 0x00);
	if (ret < 0)
		dev_err(mc->dev, "Couldn't enable standby mode\n");
	
	return ret;
}

static int mc3419_wake(struct mc3419_data *mc) {
	int ret = 0;

	ret = mc3419_write_reg(mc, MC3419_REG_MODE, 0x01);

	if (ret < 0)
		dev_err(mc->dev, "Couldn't enable wake mode\n");
	
	return ret;
}

static int mc3419_set_anymotion_threshold(struct mc3419_data *mc, uint16_t threshold) {
	int ret = 0;

	dev_dbg(mc->dev, "Set anymotion thresold to %d\n", threshold);

	uint8_t thrs[2] = {0};

	thrs[0] = (threshold & 0x00ff);
	thrs[1] = ((threshold & 0x7f00) >> 8);

	ret = mc3419_write_reg(mc, MC3419_REG_ANYM_TRESHOLD_1, thrs[0]);
	if (ret < 0)
		dev_err(mc->dev, "Error writing threshold 0\n");

	ret = mc3419_write_reg(mc, MC3419_REG_ANYM_TRESHOLD_2, thrs[1]);
	if (ret < 0)
		dev_err(mc->dev, "Error writing threshold 1\n");

	return ret;
}

static int mc3419_enable_anymotion(struct mc3419_data *mc) {
	int ret = 0;

	// set standby for register edit
	ret = mc3419_standby(mc);
	if (ret < 0)
		return ret;

	// set motion threshold
	mc3419_set_anymotion_threshold(mc, mc->anymotion_threshold);

	// enable anymotion
	ret = mc3419_write_reg(mc, MC3419_REG_MOTION_CTRL, 0x44);
	if (ret < 0) {
		dev_err(mc->dev, "Couldn't enable anymotion\n");
		return ret;
	}

	dev_dbg(mc->dev, "Anymotion enabled\n");
		
	// enable anymotion interrupt
	ret = mc3419_write_reg(mc, MC3419_REG_INT_ENABLE, 0x44);
	if (ret < 0) {
		dev_dbg(mc->dev, "Couldn't enable anymotion interrupt\n");
		return ret;
	}

	dev_dbg(mc->dev, "Anymotion interrupt enabled\n");

	// set INT2 gpio to push pull
	ret = mc3419_write_reg(mc, MC3419_REG_GPIO, 0x88);
	if (ret < 0) {
		dev_dbg(mc->dev, "Couldn't setup gpio for interrupt\n");
		return ret;
	}

	dev_dbg(mc->dev, "Gpio setup for interrupt\n");

	// swap INT1 and INT2
	ret = mc3419_write_reg(mc, MC3419_REG_INT_SWAP, 0x10);
	if (ret < 0) {
		dev_dbg(mc->dev, "Couldn't swap interrupts\n");
		return ret;
	}

	// enable wake mode for normal operation
	ret = mc3419_wake(mc);
	if (ret < 0)
		return ret;

	return ret;
}

static int mc3419_set_range(struct mc3419_data *mc) {
	int ret = 0;

	mc3419_standby(mc);

	unsigned int value;

	ret = mc3419_read_reg(mc, MC3419_REG_RANGE, &value);
	if (ret < 0)
		dev_err(mc->dev, "Failed to read range register\n");

	dev_dbg(mc->dev, "Current range: %i\n", value);

	ret = mc3419_write_reg(mc, MC3419_REG_RANGE, 0x08);
	if (ret < 0)
		dev_err(mc->dev, "Failed to write range register\n");

	return ret;
}

static int mc3419_set_sample_rate(struct mc3419_data *mc) {
	int ret = 0;

	mc3419_standby(mc);

	unsigned int value;

	ret = mc3419_read_reg(mc, MC3419_REG_SAMPLE_RATE, &value);
	if (ret < 0)
		dev_err(mc->dev, "Failed to read sample rate register\n");

	dev_dbg(mc->dev, "Current sample rate: %i\n", value);

	ret = mc3419_write_reg(mc, MC3419_REG_SAMPLE_RATE, 0x10);
	if (ret < 0)
		dev_err(mc->dev, "Failed to write sample rate register\n");

	return ret;
}

static int mc3419_reset(struct mc3419_data *mc) {
	int ret = 0;

	ret = mc3419_write_reg(mc, MC3419_REG_INT_ENABLE, 0x00);
	if (ret < 0)
		dev_err(mc->dev, "Disable interrupt failed\n");

	ndelay(10);

	return ret;
}

static int mc3419_start(struct mc3419_data *mc) {
	int ret = 0;

	// reset
	ret = mc3419_reset(mc);
	if (ret < 0)
		return ret;

	// set standby
	ret = mc3419_standby(mc);
	if (ret < 0)
		return ret;

	// set range
	ret = mc3419_set_range(mc);
	if (ret < 0)
		return ret;

	// set sample rate
	ret = mc3419_set_sample_rate(mc);
	if (ret < 0)
		return ret;

	// set wake
	ret = mc3419_wake(mc);
	if (ret < 0)
		return ret;

	return ret;
}

static ssize_t mc3419_anymotion_threshold_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct mc3419_data *mc = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", mc->anymotion_threshold);
}

static ssize_t mc3419_anymotion_threshold_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
	struct mc3419_data *mc = dev_get_drvdata(dev);

	kstrtoint(buf, 10, &mc->anymotion_threshold);
	return count;
}

static DEVICE_ATTR(anymotion_threshold, S_IRUSR | S_IWUSR, mc3419_anymotion_threshold_show, mc3419_anymotion_threshold_store);

static ssize_t mc3419_interrupt_enabled_show(struct device *dev, struct device_attribute *attr, char *buf) {
	struct mc3419_data *mc = dev_get_drvdata(dev);

	return sprintf(buf, "%d\n", mc->interrupt_enabled);
}

static ssize_t mc3419_interrupt_enabled_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
	struct mc3419_data *mc = dev_get_drvdata(dev);

	kstrtoint(buf, 10, &mc->interrupt_enabled);

	if (mc->interrupt_enabled == 1) {
		mc3419_start(mc);
		mc3419_enable_anymotion(mc);
	} else {
		mc3419_reset(mc);
		mc3419_standby(mc);
	}

	return count;
}

static DEVICE_ATTR(interrupt_enabled, S_IRUSR | S_IWUSR, mc3419_interrupt_enabled_show, mc3419_interrupt_enabled_store);


static int mc3419_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct device_node *np = client->dev.of_node;
	struct mc3419_data *mc;
	int ret = 0;

	dev_dbg(&client->dev, "Probing\n");

	if (!(IS_ENABLED(CONFIG_OF) && np)) {
		dev_err(&client->dev,
			"No device tree data for '%s'! Platform data not supported.\n",
			client->name);
		return -EINVAL;
	}

	mc = devm_kzalloc(&client->dev, sizeof(*mc), GFP_KERNEL);
	if (!mc)
		return -ENOMEM;

	mc->client = client;
	mc->dev	   = &client->dev;
	mc->regmap = devm_regmap_init_i2c(client, &mc3419_regmap_config);
	if (IS_ERR(mc->regmap)) {
		dev_err(mc->dev, "regmap initialization failed.\n");
		devm_kfree(mc->dev, mc);
		return PTR_ERR(mc->regmap);
	}

	i2c_set_clientdata(client, mc);

	// add  input device
	mc->input = input_allocate_device();
	if (!mc->input) {
		dev_err(mc->dev, "Failed to allocate input device\n");
	}

	mc->input->name = "mc3419-accelerometer";
	mc->input->evbit[0] = BIT_MASK(EV_KEY);
	mc->input->keybit[BIT_WORD(KEY_WAKEUP)] = BIT_MASK(KEY_WAKEUP);

	ret = input_register_device(mc->input);
	if (ret) {
		dev_err(mc->dev, "Failed to register input device\n");
		input_free_device(mc->input);
	}

	dev_dbg(mc->dev, "Initializing...\n");

	if (client->irq) {
		dev_dbg(mc->dev, "Setting up irq handling...\n");

		ret = devm_request_threaded_irq(&client->dev, client->irq, NULL, mc3419_irq_handler, IRQF_TRIGGER_FALLING | IRQF_ONESHOT, devm_kasprintf(mc->dev, GFP_KERNEL, "mc3419.%s", dev_name(&client->dev)), mc);
		if (ret) {
			dev_err(&client->dev, "Unable to register IRQ %d error %d\n", client->irq, ret);
			return ret;
		}
	}

	// create sysfs groups
	ret = device_create_file(&client->dev, &dev_attr_anymotion_threshold);
	if (ret) {
		dev_err(&client->dev,
			"Sysfs create failure anymotion_threshold.\n");
	}

	mc->anymotion_threshold = 120;

	ret = device_create_file(&client->dev, &dev_attr_interrupt_enabled);
	if (ret) {
		dev_err(&client->dev,
			"Sysfs create failure interrupt_enabled.\n");
	}

	mc->interrupt_enabled = 1;

	// init the chip
	mc3419_start(mc);
	mc3419_enable_anymotion(mc);

	dev_dbg(mc->dev, "All good\n");

	return 0;
}

static void mc3419_remove(struct i2c_client *client) {
	struct mc3419_data *mc = i2c_get_clientdata(client);

	device_remove_file(&client->dev, &dev_attr_anymotion_threshold);
	device_remove_file(&client->dev, &dev_attr_interrupt_enabled);

	input_unregister_device(mc->input);

	devm_kfree(&client->dev, mc);
}

#if defined(CONFIG_PM_SLEEP)
static int mc3419_suspend(struct device *dev) {
	struct i2c_client *client = to_i2c_client(dev);
	struct mc3419_data *mc = i2c_get_clientdata(client);

	// // start the chip
	// mc3419_start(mc);

	// // enable motion detect
	// mc3419_enable_anymotion(mc);

	if (mc->interrupt_enabled) {
		disable_irq(client->irq);

		enable_irq_wake(client->irq);
	}
	
	return 0;
}

static int mc3419_resume(struct device *dev) {
	struct i2c_client *client = to_i2c_client(dev);
	struct mc3419_data *mc = i2c_get_clientdata(client);

	if (mc->interrupt_enabled) {
		enable_irq(client->irq);

		disable_irq_wake(client->irq);
	}

	// // reset
	// mc3419_reset(mc);

	// // set standby
	// mc3419_standby(mc);

	return 0;
}
#endif

static SIMPLE_DEV_PM_OPS(mc3419_pm_ops, mc3419_suspend, mc3419_resume);

static struct i2c_driver mc3419_driver = {
	.driver = {
		.name = "mc3419",
		.pm = &mc3419_pm_ops,
		.of_match_table = of_match_ptr(mc3419_of_table),
	},
	.probe    = mc3419_probe,
	.remove	  = mc3419_remove,
	.id_table = mc3419_id,
};

module_i2c_driver(mc3419_driver);

MODULE_AUTHOR("Marton Borzak <marton@unfoldedcircle.com>");
MODULE_DESCRIPTION("Driver for MC3419 accelerometer");
MODULE_LICENSE("GPL");
