// SPDX-License-Identifier: GPL-2.0-only
/*
Driver for testing ADC on UC Remote TWo.
Just a quick and dirty thing to test the IR phototransistor is working.

Device tree entry:

adc_ir_receive {
       status = "okay";
       compatible = "uc,adc-ir-receive";
       io-channels = <&saradc 2>; // saradc channel
};

*/
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/of_gpio.h>
#include <linux/gpio.h>

#include <linux/iio/iio.h>
#include <linux/iio/machine.h>
#include <linux/iio/driver.h>
#include <linux/iio/consumer.h>

#define ADC_SAMPLE_JIFFIES      (1 / (MSEC_PER_SEC / HZ))
#define VREF 1800

struct iio_channel *channel;
struct delayed_work adc_poll_work;
int start_test;
int test_result;

static void uc_adc_test_poll(struct work_struct *work)
{
    int ret,raw;
    int result = -1;

    if (channel == NULL) {
        return;
    }

    ret = iio_read_channel_raw(channel, &raw);
    if (ret < 0) {
        printk("error: %d\n", ret);
        return;
    }

    if (raw < 900) {
        result = (VREF*raw)/1023;
        // printk("ADC raw= %d Voltage= %dmV\n", raw, result);
    }

    if (result > 0 && result < 1200) {
        test_result = 0;
        // printk("Test result OK");
    }

    if (start_test == 1)
        schedule_delayed_work(&adc_poll_work, ADC_SAMPLE_JIFFIES);
}

static ssize_t uc_adc_start_test_show(struct device *dev, struct device_attribute *attr, char *buf) {
	return sprintf(buf, "%d\n", start_test);
}

static ssize_t uc_adc_start_test_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
	kstrtoint(buf, 10, &start_test);

    if (start_test == 1) {
        test_result = -1;
        schedule_delayed_work(&adc_poll_work, ADC_SAMPLE_JIFFIES);
    }

	return count;
}

static DEVICE_ATTR(start_test, S_IRUSR | S_IWUSR, uc_adc_start_test_show, uc_adc_start_test_store);


static ssize_t uc_adc_test_result_show(struct device *dev, struct device_attribute *attr, char *buf) {
	return sprintf(buf, "%d\n", test_result);
}

static ssize_t uc_adc_test_result_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count) {
	kstrtoint(buf, 10, &test_result);
	return count;
}

static DEVICE_ATTR(test_result, S_IRUSR | S_IWUSR, uc_adc_test_result_show, uc_adc_test_result_store);


static int uc_adc_probe(struct platform_device *pdev)
{
    channel = iio_channel_get(&(pdev->dev), NULL);
    if (IS_ERR(channel)) {
        channel = NULL;
        dev_err(&pdev->dev, "missing adc channel\n");
        return -1;
    }

    if (channel) {
        INIT_DELAYED_WORK(&adc_poll_work, uc_adc_test_poll);
        schedule_delayed_work(&adc_poll_work,1000);
    } else {
        dev_err(&pdev->dev, "adc channel is null\n");
        return -1;
    }

	int ret = device_create_file(&pdev->dev, &dev_attr_start_test);
	if (ret) {
		dev_err(&pdev->dev,
			"Sysfs create failure start_test.\n");
	}

    start_test = 0;

    ret = device_create_file(&pdev->dev, &dev_attr_test_result);
	if (ret) {
		dev_err(&pdev->dev,
			"Sysfs create failure test_result.\n");
	}

    test_result = -1;

    dev_info(&pdev->dev, "UC ADC test probe!\n");

    return 0;
}

static int uc_adc_remove(struct platform_device *pdev)
{
    if (channel)
        iio_channel_release(channel);

    channel = NULL;
        
    device_remove_file(&pdev->dev, &dev_attr_start_test);
    device_remove_file(&pdev->dev, &dev_attr_test_result);
    return 0;
}

static const struct of_device_id uc_adc_match[] = {
    { .compatible = "uc,adc-ir-receive" },
    {},
};

static struct platform_driver uc_adc_driver = {
    .probe      = uc_adc_probe,
    .remove     = uc_adc_remove,
    .driver     = {
        .name   = "uc_adc_ir_receive",
        .owner  = THIS_MODULE,
        .of_match_table = uc_adc_match,
    },
};

static int uc_adc_init(void)
{
    return platform_driver_register(&uc_adc_driver);
}
late_initcall(uc_adc_init);

static void uc_adc_exit(void)
{
    platform_driver_unregister(&uc_adc_driver);
}
module_exit(uc_adc_exit);

MODULE_AUTHOR("Marton Borzak <marton@unfoldedcircle.com>");
MODULE_DESCRIPTION("UC ADC IR receive driver");
MODULE_ALIAS("platform:uc-adc-ir-receive");
MODULE_LICENSE("GPL");
